
=head1 NAME

ll3model.pm

=head1 SYNOPSIS

This module can be used calculate and return the log likelihood 
values based on the four expected value assumptions:
    1) P(xyz) = P(x)P(y)P(z)
    2) P(xyz) = P(xy)P(z)
    3) P(xyz) = P(x)P(yz)
    4) P(xyz) = P(xz)P(y)

In particular this module provides functions that give convenient access
to 3-d (i.e., trigram) frequency counts as created by count.pl, as well 
as some degree of error handling that verifies the data.

To be used in a measure module that is to be used by statistic.pl, the
functions provided in this module must be embedded within other functions
that adhere to the standards and naming convention described in
Docs/NewStats.txt and is briefly summarized below.

=head1 DESCRIPTION

The functions in this module retrieve observed trigram frequency counts 
and marginal totals, compute expected values and the log likelihood scores
for each of the possible assumptions. They also provide support for error 
checking of the output produced by count.pl. These functions are used in 
all the trigram (3d) measure modules.

The Log Likelhood measure is the ratio between how often an ngram actually 
occurred compared to how often it would be expected occur. In this measure, 
you are comparing the observed and expected values for each "cell" in a 
contingency table that represents the counts of your observed ngram. 

In the 2 dimensional case, there is only one possible model that 
you can compare to, and that is the model of independence. The log 
likelihood calculates the observed count of your bigram, and compares 
it to the counts that would be expected if the words in the bigram were 
statistically independent. Statistical independence is the probability 
that two words occur together is equal to the product of the their 
individual probabilites. More formally, this can be defined as 
P(word1, word2) = P(word1) * P(word2)

The score that log-likelihood measure produces reflects the degree to 
which the observed and expected values diverge. A log likelihood score 
of zero implies that the data fits perfectly into the independence model 
and the expected values are equal to the observed values. Therefore, the 
higher the log likelihood score, the less the tokens in the ngram 
appear to be independent from each other. 

Calculating the expected values based on independence can be carried to 
trigrams but as the dimensions of your contingency table grow, so does the 
number of models in which the words can be compared to. The expected values 
for a 3grams can be based on 4 other models listed below:
 
Model 1: P(w1 w2 w3) = P(w1) P(w2) P(w3)
Model 2: P(w1 w2 w3) = P(w1 w2) P(w3)
Model 3: P(w1 w2 w3) = P(w1) P(w2 w3)
Model 4: P(w1 w2 w3) = P(w1 w3) P(w2) 

Each of these models will result in different expected values and 
log likelihood scores. With 3gram or 3d measures we use a 3x3 contingency 
table to store the  frequency counts associated with each word in the 4gram, 
as well as the  number of times the 4gram occurs. The notation we employ is 
as follows:

Marginal Frequencies:
    
 n1pp = the number of trigrams where the first word is word1.
 np1p = the number of trigrams where the second word is word2.
 npp1 = the number of trigrams where the third word is word3.
 n2pp = the number of trigrams where the first word is not word1.
 np2p = the number of trigrams where the second word is not word2.
 npp2 = the number of trigrams where the third word is not word3.

Observed Frequencies:

 n111 = number of times word1, word2 and word3 occur together in
        their respective positions, joint frequency.
 n112 = number of times word1 and word2 occur in their respective
        positions but word3 does not.
 n211 = number of times word2 and word3 occur in their respective
        positions but word1 does not.
 n212 = number of times word2 occurs in in its respective position
        but word1 and word2 do not.
 n121 = number of times word1 and word3 occur in their respective
        positions but word2 does not.
 n122 = number of times word1 occurs in its respective position
        but word2 and word3 do not.
 n221 = number of times word3 occurs in its respective position
        but word1 and word2 do not.
 n222 = number of time neither word1, word2 or word3 occur in their
        respective positions.

Expected Frequencies (Examples for the Independence Assumption) :

 m111 = expected number of times word1, word2 and word3 occur together in
        their respective positions.
 m112 = expected number of times word1 and word2 occur in their respective
        positions but word3 does not.
 m211 = expected number of times word2 and word3 occur in their respective
        positions but word1 does not.
 m212 = expected number of times word2 occurs in in its respective position
        but word1 and word2 do not.
 m121 = expected number of times word1 and word3 occur in their respective 
        positions but word2 does not.
 m122 = expected number of times word1 occurs in its respective position
        but word2 and word3 do not.
 m221 = expected number of times word3 occurs in its respective position
        but word1 and word2 do not.
 m222 = expected number of time neither word1, word2 or word3 occur in their
        respective positions.

=head2 Functions Included in ll3model

=over 4

=item 1 

The intializeStatistic function performs initialization before
the actual computation of measures of association begin. It also
verifies that the input consists of trigrams, and that it has all
the frequency combinations needed to do the computations. The
frequency combinations that are required for this module are
$n111, $n1pp, $np1p, $npp1, $n11p, $np11, and $n1p1.

initializeStatistic() is passed the following input parameters:

     1) The ngram size. For 3d (trigram) measures this will be 3. 
     2) The total number of trigrams in the corpus (nppp).
     3) The number of frequency combinations. 
     4) A 3-d array containing the frequency combinations.

Each row  of the array in 4) represents a single frequency combination.
On a given row, the first element denotes the number of indices on this
row, say 'n'. This is followed by the 'n' values that correspond to the
indices included in the frequency combination. (For more details on
frequency combinations, see README.pod). To use this module, the joint
frequency, as well as the marginal frequencies are required in order to
calculate the expected values.

This function does not return any values. If an error occurs, it can be
detected by statistic.pl using the errorCode and errorString functions
described below.

=item 2

The checkObservedValues function takes as input an array containing the
frequency values for a trigram as found by count.pl. This will include
seven values: n111, n1pp, np1p, npp1, n11p, np11, and n1p1. The size
of this array is guaranteed to be exactly the same as the third parameter
passed to the function initializeStatistic() function above.

The checkObservedValues function verifies that the marginal frequencies
n1pp, np1p, npp1, n11p, np11, and n1p1 are consistent with the value
of the joint frequency of the trigram, n111. If they are not consistent,
it sets an error code and error message, and the function returns.

If the marginals are valid, it computes the observed values for the
remaining cells in the 3d table ( n112, n211, n212, n121, n122, n221,
n222 ) based on these marginal totals and the joint frequency and
returns 1.

=item 3

The function checkExpectedValues() calculates the expected values of the 
cells in the contingency table based on the marginal frequencies and the 
total sample size. The expected values are estimated based on each of the
assumption  models as well as the loglieklihood scores for each set of the
expected values. The function returns a 1 if everything is okay.

=item 4

The function getMarginalValues will return the marginal frequencies in the
order of n1pp, np1p, npp1, n11p, np11, n1p1, n2pp, np2p, npp2.

=item 5

The function getTotalTrigrams() returns the total number of trigrams in the
corpus (nppp).

= item 6

The function getLogLikelihoodValues() returns the log likelihood scores
for each of the assumption models in the following order: $LogLikelihood1,  
$LogLikelihood2, $LogLikelihood3, $LogLikelihood4 where each of the 
numbers cooresponds to the above described models.

=item 6

The function errorCode() returns 0 if the last operation was successful.
It will return an integer starting with 1 if the last operation failed.
This indicates that statistic.pl should abort. It will return an integer
starting with 2 to indicate a warning should be issued. This does not
cause cause statistic.pl to abort. However, a warning after
calculateStatistic() will cause the trigram which generated that warning to
be ignored by statistic.pl.

=item 7

The function errorString() returns the text of an error message.

=back

=head2 Writing your own statistic model module

Any measure module being used by statistic.pl must follow this
convention. In order to make it easier to build 3d measures, we provide
a 3d specific functions that can be embedded within the measure module in
order to carry out common operations in calculating the values of such
measures.

=over 4

=item 1

The filename should have an extension of .pm. Usually the name of
the file should be Statistic.pm, where "Statistic" is the name of the
particular statistic you are writing.

=item 2

Let us say you have named your file Statistic.pm. The first line of the
file should declare that its a package of the same name as the filename.
Thus the first line of the file Statistic.pm should be:

   package Statistic;

=item 3

To use the ll3model.pm module you need to include it in your
Statistic.pm module.

A small code snippet to ensure that it is included is as follows:

 #  Check to see if the statistic.pm module can see the
 #  ll3model.pm module ... if not see if it can be found.

 my $module = "ll3model.pm"; my $modulename = "ll3model.pm";
 if( !( -f $modulename ) ) {
    my $found = 0;
    #  Check each of the PATHS to see if the module is there
    foreach (split(/:/, $ENV{PATH})) {
 	$module = $_ . "/" . $modulename;
	if ( -f $module ) { $found = 1; last; }
    }
    # if still not found anywhere, quit!
    if ( ! $found ) { print "Could not find $modulename.\n"; exit; }
 }

 # IMPORTANT : now include the module into the current package    
 require $module;

=item 4 

You need to implement at least two functions in your package

   i)   initializeStatistic()
   ii)  calculateStatistic()


Function initializeStatistic() is passed the a set of parameters
that include the trigram size which is 3, the total number of
trigrams in the corpus, the number of frequency combinations
and an array containing the frequency combinations. More detail
of these parameters are described above in the description of
the ll3model initializeStatistic function.

These paramters can be passed directly into the ll3model.pm
modules intializeStatistic. For example:

 sub initializeStatistic {

     ll3model::initializeStatistic(@_);

 }

where @_ contains the input parameters.

This function is called before any calls to the function
calculateStatistic() and can be used by the statistic library
to set up any values that may be required for the calculations
later. This function is not expected to return anything. If an
error occurs, it can be reported through the mechanisms described 
below.

The other mandatory function is calculateStatistic(). This is
passed an array containing the frequency values for an ngram as
found in the input n-gram file.

Function calculateStatistic() is expected to return a (possibly
floating) value as the value of the statistical measure calculated
using the frequency values passed to it.

There exists two main functions in the module ll3model.pm in order
to help calculate the trigram statistic.

   1.  ll3model::checkObservedValues(@frequencies)
   2.  ll3model::checkExpectedValues();

The function checkObservedValues will return a 1. If it does not then 
there existed an error in the calculation of these values and zero should 
be returned. An example of how this can be used is as follows:
  
if( !( $check = ll3model::getObservedValues(@_) ) ) {
	return(0);
}

where @_ is the parameters sent to calculateStatistic from statistic.pl.
A more detailed description of this function can be seen above.

The function getExpectedValues will return a 1. If it does not then there 
existed an error in the calculation of these values and zero should be 
returned. An example of how this can be used is as follows:

if( !( $check = ll3model::getExpectedValues() ) ) {
    return(0);
}

The function ll4model::checkExpectedValues will return a 1.
If it does not then there existed an error in the calculation of 
these values and zero should be returned. An example of how this
can be used is as follows:

When a library is loaded, statistic.pl checks for initializeStatistc
and calcualteStatistic functions: if they are not implemented, then
an error is reported and the program quits.

The function getLogLikelihoodValues will return a list of the
log likelihood scores for each of the assumption models. An
example of how htis can be used is as follows:

if( ! ( ($LogLikelihood1,  $LogLikelihood2, 
         $LogLikelihood3,  $LogLikelihood4) 
         = ll4model::getLogLikelihoodValues())) {
    
    return(0);
}

=item 5 

Program statistic.pl also supports three other functions that are
not mandatory, but may be implemented by the user. These are:

     i) errorCode()
    ii) errorString()
   iii) getStatisticName()

Function errorCode, if implemented, is called immediately after
the call to function initializeStatistic() and immediately after
every call to function calculateStatistic().

The ll3model.pm module implements both errorCode()
and errorString().

The errorCode() and errorString() methods that are implmented in
your Statistic.pm modoule can return the value returned by the
ll3model errorCode() and errorString() functions.

An example of this is below:

sub errorCode
{
  return ll3model::errorCode();
}

sub errorString
{
    return ll3model::errorString();
}

The third function that may be implemented is getStatisticName().
If this function is implemented, it is expected to return a string
containing the name of the statistic being implmented. This string
is used in the formatted output of statistic.pl. If this function
is not implemented, then the statistic file name entered on the
commandline is used in the formatted output.

Note that all three functions described in this section are first
checked for existence before being called. So, if the user elects
to not implement these functions, no harm will be done. However, we
strongly recommend the implementation of at least the function
errorCode() since this is the only way for the statistic library to
report errors to the user.

=item 6

Having implemented the two mandatory functions (in point 3 above)
and zero or more of the three non-mandatory functions (in point 4
above), one must make these functions available outside the
package. To do so, one has to export them, thusly.

For this, first include the Exporter package by including the
following line in the program

   require Exporter;

Now include the following line to inherit Exporters functions:

   @ISA = qw ( Exporter );

Now export the various functions implemented so that they are
accessible outside this package, by adding the following line
(assume that you have implemented only the two mandatory
functions):

   @EXPORT = qw( initializeStatistic calculateStatistic );

If you implement say the errorCode() and errorString() functions
too, you may export them like so:

   @EXPORT = qw( initializeStatistic calculateStatistic errorCode errorString );
   
Note that the user may implement other functions too, and may
export them if he so wishes, but since statistic.pl is not
expecting anything besides the five functions above, doing so would
have no effect on statistic.pl.

=item 7 

Finally, at the end of everything, add the line

   1;

   This will ensure that the LAST line of the file returns a 
   true value, and is necessary so that when this package is 
   loaded, it returns a TRUE value.

=back

=head2 Errors to look out for:

=over 4

=item 1 

The filename does not end with a .pm.

=item 2 

The rest of the filename (besides the extension) does not match the package
name (declared in the first line of the file). Remember its case sensitive!

=item 3 

The five functions (2 mandatory, 3 non-mandatory) must have their
names match EXACTLY with those shown above. Again, names are all
case sensitive.

=item 4 

The last line of the file is not "1;". This is necessary, and easily
overlooked!

=back

=head1 AUTHORS

    Ted Pedersen (tpederse@umn.edu)
    Satanjeev Banerjee <banerjee@cs.cmu.edu>
    Bridget McInnes (bthomson@d.umn.edu)

=head1 BUGS

=head1 SEE ALSO

 home page:    http://www.d.umn.edu/~tpederse/nsp.html

 mailing list: http://groups.yahoo.com/group/ngram/

=head1 COPYRIGHT

Copyright (C) 2004 Satanjeev Banerjee, Ted Pedersen and Bridget McInnes

Permission is granted to copy, distribute and/or modify this  documeny
under the terms of the GNU Free Documentation License, Version 1.2 or
any later version published by the Free Software Foundation; with no
Invariant Sections, no Front-Cover Texts, and no Back-Cover Texts.

Note: a copy of the GNU Free Documentation License is available on the
web at L<http://www.gnu.org/copyleft/fdl.html> and is included in this
distribution as FDL.txt. 

=cut
    
package ll3model;
require Exporter;

@ISA = qw ( Exporter );
@EXPORT = qw( initializeStatistic checkObservedValues checkExpectedValues getLogLikelihoodValues getMarginalTotals getTotalTrigrams errorCode errorString );


# function to set up various variables before the actual computation
# starts. also to check if we are being given trigrams, and if our
# frequency combinations are enough to do the computation

sub initializeStatistic
{
    ($ngram, $nppp, $combIndex, @freqComb) = @_;
    
    $errorCodeNumber = 0; $errorMessage = "";
    
    # check if ngram !=3 . statistic only defined for ngram = 3. 
    if ($ngram != 3) {
	$errorMessage = "This statistic is only available for trigrams!";
	$errorCodeNumber = 1;  return;
    }
    
    # The total number of trigrams should not be less than equal to 0
    if ($nppp <= 0) { 
	$errorMessage = "Total number of trigrams ($nppp) must be greater than 0.";
	$errorCodeNumber = 10;  return;
    }
    
    # figure out from the @freqComb array if the frequency values we
    # are going to get are indeed the ones we need. 
    # while we figure this out, we shall also note which of the indices of the 
    # array passed to function calculateStatistic are the ones we want.
    
    for (my $i = 0; $i < $combIndex; $i++)
    {
	$str="";
        foreach(@{$freqComb[$i]}[1..$freqComb[$i][0]]) { $str.=$_."#"; }
        if($str eq "0#1#2#") {  $n111Index=$i; }
        if($str eq "0#")     {  $n1ppIndex=$i; }
        if($str eq "1#")     {  $np1pIndex=$i; }
	if($str eq "2#")     {  $npp1Index=$i; }
        if($str eq "0#1#")   {  $n11pIndex=$i; }
        if($str eq "1#2#")   {  $np11Index=$i; }
        if($str eq "0#2#")   {  $n1p1Index=$i; }
    }
        
    if (!(defined $n111Index)) { $errorCodeNumber = 100; $errorMessage = "Frequency combination \"0 1 2\" missing!\n"; }
    if (!(defined $n1ppIndex)) { $errorCodeNumber = 101; $errorMessage = "Frequency combination \"0\" missing!\n";   }
    if (!(defined $np1pIndex)) { $errorCodeNumber = 102; $errorMessage = "Frequency combination \"1\" missing!\n";   }
    if (!(defined $npp1Index)) { $errorCodeNumber = 103; $errorMessage = "Frequency combination \"2\" missing!\n";   }
    if (!(defined $n11pIndex)) { $errorCodeNumber = 104; $errorMessage = "Frequency combination \"0 1\" missing!\n"; }
    if (!(defined $np11Index)) { $errorCodeNumber = 105; $errorMessage = "Frequency combination \"1 2\" missing!\n"; }
    if (!(defined $n1p1Index)) { $errorCodeNumber = 106; $errorMessage = "Frequency combination \"0 2\" missing!\n"; }
    
}

# function to ensure the marginal values are reasonable with
# respect to each other and if so calculate the observed values
sub checkObservedValues() 
{
    $n111=$_[$n111Index];
    $n1pp=$_[$n1ppIndex];
    $np1p=$_[$np1pIndex];
    $npp1=$_[$npp1Index];
    $n11p=$_[$n11pIndex];
    $np11=$_[$np11Index];
    $n1p1=$_[$n1p1Index];
    
    $n112=$n11p-$n111;
    $n211=$np11-$n111;
    $n212=$np1p-$n111-$n112-$n211;
    $n121=$n1p1-$n111;
    $n122=$n1pp-$n111-$n112-$n121;
    $n221=$npp1-$n111-$n211-$n121;
    $n222=$nppp-($n111+$n112+$n121+$n122+$n211+$n212+$n221);
    
    $n2pp=$nppp-$n1pp;
    $np2p=$nppp-$np1p;
    $npp2=$nppp-$npp1;
    
    # n111 should be greater than equal to zero 
    if ($n111< 0)    {
	$errorMessage = "Frequency value ($n111) must not be negative.";
	$errorCodeNumber = 200;  return;
    }
    
    # n111 frequency should be less than or equal to totalTrigrams
    if ($n111> $nppp) {
	$errorMessage = "Frequency value ($n111) must not exceed total number of trigrams.";
	$errorCodeNumber = 201;  return;
    }
    
    # joint frequency n111 should be less than or equal to the marginal totals
    if ($n111 > $n1pp || $n111 > $np1p || $n111 > $npp1) {
	$errorMessage = "Frequency value of ngram ($n111) must not exceed the marginal totals.";
	$errorCodeNumber = 202;  return;
    }
    
    # n1pp should be greater than or equal to zero 
    if ($n1pp< 0) {
	$errorMessage = "Marginal total value ($n1pp) must not be negative.";
	$errorCodeNumber = 203;  return;
    }

    # n1pp should be less than or equal to totalTrigrams
    if ($n1pp > $nppp) {
	$errorMessage = "Marginal total value ($n1pp) must not exceed total number of trigrams.";
	$errorCodeNumber = 204;  return;
    }

    # np1p should be greater than or equal to zero
    if ($np1p< 0) {
	$errorMessage = "Marginal total value ($np1p) must not be negative.";
        $errorCodeNumber = 205;  return;
    }

    # np1p should be less than or equal to totalTrigrams
    if ($np1p > $nppp) {
	$errorMessage = "Marginal total value ($np1p) must not exceed total number of trigrams.";
        $errorCodeNumber = 206;  return;
    }
    
    # npp1 should be greater than or equal to zero 
    if ($npp1< 0) {
	$errorMessage = "Marginal total value ($npp1) must not be negative.";
	$errorCodeNumber = 207;  return;
    }
    
    # npp1 should be less than or equal to totalTrigrams
    if ($npp1 > $nppp) {
	$errorMessage = "Marginal total value ($npp1) must not exceed total number of trigrams.";
	$errorCodeNumber = 208;  return;
    }

    #  Everything looks good so we can return the observed values
    @values =  ( $n111, $n112, $n121, $n122, $n211, $n212, $n221, $n222 );
     
    return 1;
}

#  Calculates, checks and returns the expected values
sub checkExpectedValues
{
    #  initialize the logLikelihood values
    $logLikelihood1 = 0; $logLikelihood2 = 0; $logLikelihood3 = 0; $logLikelihood4 = 0;


    #  based on the P(xyz) / ( P(xy) P(z) )

    $n12p = $n1pp - $n11p; $n21p = $np1p - $n11p; $n22p = $n221 + $n222;
    
    $m111=$n11p*$npp1/($nppp);	$m112=$n11p*$npp2/($nppp);
    $m121=$n12p*$npp1/($nppp);	$m122=$n12p*$npp2/($nppp);
    $m211=$n21p*$npp1/($nppp);	$m212=$n21p*$npp2/($nppp);
    $m221=$n22p*$npp1/($nppp);	$m222=$n22p*$npp2/($nppp);
    
    $logLikelihood1 = 0;
    
    #  Check marginals
    $check = CheckMarginals(); if($check == 1) { return; }

    #  Calculate the Log Likelihood    
    $logLikelihood1 = getLogLikelihood();
    
    #  based on the P(xyz) / ( P(xz) P(y) )
    
    $n1p2 = $n1pp - $n1p1;    $n2p1 = $npp1 - $n1p1;    $n2p2 = $n222 + $n212;
    
    $m111=$n1p1*$np1p/($nppp**1);	$m112=$n1p2*$np1p/($nppp**1);
    $m121=$n1p1*$np2p/($nppp**1);	$m122=$n1p2*$np2p/($nppp**1);
    $m211=$n2p1*$np1p/($nppp**1);	$m212=$n2p2*$np1p/($nppp**1);
    $m221=$n2p1*$np2p/($nppp**1);	$m222=$n2p2*$np2p/($nppp**1);
    
    #  Check marginals
    $check = CheckMarginals(); if($check == 1) { return; }

    #  Calculate the Log Likelihood    
    $logLikelihood2 = getLogLikelihood();

  
    #  based on the P(xyz) / ( P(x) P(yz) )

    $np12 = $np1p - $np11;    $np21 = $npp1 - $np11;    $np22 = $n222 + $n122;
    
    $m111=$n1pp*$np11/($nppp**1);	$m112=$n1pp*$np12/($nppp**1);
    $m121=$n1pp*$np21/($nppp**1);	$m122=$n1pp*$np22/($nppp**1);
    $m211=$n2pp*$np11/($nppp**1);	$m212=$n2pp*$np12/($nppp**1);
    $m221=$n2pp*$np21/($nppp**1);	$m222=$n2pp*$np22/($nppp**1);  
    
    #  Check marginals
    $check = CheckMarginals(); if($check == 1) { return; }
    
    #  Calculate the Log Likelihood    
    $logLikelihood3 = getLogLikelihood();
    
    
    #  Calculate the expected values based on the model of independence   

    $m111=$n1pp*$np1p*$npp1/($nppp**2); $m112=$n1pp*$np1p*$npp2/($nppp**2);
    $m121=$n1pp*$np2p*$npp1/($nppp**2); $m122=$n1pp*$np2p*$npp2/($nppp**2);
    $m211=$n2pp*$np1p*$npp1/($nppp**2); $m212=$n2pp*$np1p*$npp2/($nppp**2);
    $m221=$n2pp*$np2p*$npp1/($nppp**2); $m222=$n2pp*$np2p*$npp2/($nppp**2);
    
    #  Check marginals
    $check = CheckMarginals(); if($check == 1) { return; }
    
    #  Get the Log Likelihood    
    $logLikelihood4 = getLogLikelihood();
    
    return 1;
    
}

#  calculate the log likelihood
sub getLogLikelihood
{
    my $logLikelihood = 0;

    if ( $n111 && $m111 ) { $logLikelihood += $n111 * log ( $n111 / $m111 ); }
    if ( $n112 && $m112 ) { $logLikelihood += $n112 * log ( $n112 / $m112 ); }
    if ( $n121 && $m121 ) { $logLikelihood += $n121 * log ( $n121 / $m121 ); }
    if ( $n122 && $m122 ) { $logLikelihood += $n122 * log ( $n122 / $m122 ); }
    if ( $n211 && $m211 ) { $logLikelihood += $n211 * log ( $n211 / $m211 ); }
    if ( $n212 && $m212 ) { $logLikelihood += $n212 * log ( $n212 / $m212 ); }
    if ( $n221 && $m221 ) { $logLikelihood += $n221 * log ( $n221 / $m221 ); }
    if ( $n222 && $m222 ) { $logLikelihood += $n222 * log ( $n222 / $m222 ); }
    
    return (2 * $logLikelihood);

}


   #  return the log likelihood values
sub getLogLikelihoodValues
{
    @values = ($logLikelihood1, $logLikelihood2, $logLikelihood3, $logLikelihood4);

    return @values;

}

#  check to ensure that the marginals are accurate
sub CheckMarginals {
   
    # dont want ($n111 / $m111) to be 0 or less! flag error if so!
    if ( $n111 )  
    { 
	if ($m111 == 0) {
	    $errorMessage = "Expected value in cell (1,1,1) must not be zero";
	    $errorCodeNumber = 231;  return 1;
	}

	if (($n111 / $m111) < 0) {
	    $errorMessage = "About to take log of negative value for cell (1,1)";
	    $errorCodeNumber = 232;  return 1;
	}
    }

    if ( $n112 ) 
    { 
	if ($m112 == 0) {
	    $errorMessage = "Expected value in cell (1,1,2) must not be zero";
	    $errorCodeNumber = 233;  return 1;
	}
	
	if (($n112 / $m112) < 0) {
	    $errorMessage = "About to take log of negative value for cell (1,1,2)";
	    $errorCodeNumber = 234;  return 1;
	}
    }

    if ( $n121 ) 
    { 
	if ($m121 == 0) {
	    $errorMessage = "Expected value in cell (1,2,1) must not be zero";
	    $errorCodeNumber = 235;  return 1;
	}

	if (($n121 / $m121) < 0) {
	    $errorMessage = "About to take log of negative value for cell (1,2,1)";
	    $errorCodeNumber = 236;  return 1;
	}
    }

    if ( $n122 ) 
    { 
	if ($m122 == 0)	{
	    $errorMessage = "Expected value in cell (1,2,2) must not be zero";
	    $errorCodeNumber = 237;  return 1;
	}

	if (($n122 / $m122) < 0) {
	    $errorMessage = "About to take log of negative value for cell (1,2,2)";
	    $errorCodeNumber = 238;  return 1;
	}
    }
    
    if ( $n211 )
    {
        if ($m211 == 0) {
	    $errorMessage = "Expected value in cell (2,1,1) must not be zero";
            $errorCodeNumber = 239;  return 1;
        }
	
        if (($n211 / $m211) < 0) {
	    $errorMessage = "About to take log of negative value for cell (2,1,1)";
            $errorCodeNumber = 240;  return 1;
        }
    }
    
    if ( $n212 )
    {
        if ($m212 == 0) {
	    $errorMessage = "Expected value in cell (2,1,2) must not be zero";
            $errorCodeNumber = 241;  return 1;
        }
	
        if (($n212 / $m212) < 0) {
	    $errorMessage = "About to take log of negative value for cell (2,1,2)";
            $errorCodeNumber = 242;  return 1;
        }
    }

    if ( $n221 )
    {
        if ($m221 == 0)  {
	    $errorMessage = "Expected value in cell (2,2,1) must not be zero";
            $errorCodeNumber = 243;  return 1;
        }

        if (($n221 / $m221) < 0) {
	    $errorMessage = "About to take log of negative value for cell (2,2,1)";
            $errorCodeNumber = 244;  return 1;
        }
    }
    
    if ( $n222 )
    {
        if ($m222 == 0) {
	    $errorMessage = "Expected value in cell (2,2,2) must not be zero";
            $errorCodeNumber = 245;  return 1;
        }
	
        if (($n222 / $m222) < 0) {
	    $errorMessage = "About to take log of negative value for cell (2,2,2)";
            $errorCodeNumber = 246;  return 1;
        }
    }
     
    return 0;
}

#  Returns the marginal frequencies
sub getMarginalTotals {
    
    @values = ( $n1pp, $np1p, $npp1, $n11p, $np11, $n1p1, $n2pp, $np2p, $npp2 );
    return @values;
}

#  Return the total number of trigrams
sub getTotalTrigrams {
    return $nppp;
}

# function to return the error code of the last operation and reset
# error code. useful if the error can be recovered from!
sub errorCode 
{ 
    my $temp = $errorCodeNumber;
    $errorCodeNumber = 0;
    return($temp); 
}


# function to return the error message of the last operation and reset
# the message string. useful if error can be recovered from!
sub errorString
{
    my $temp = $errorMessage;
    $errorMessage = "";
    return($temp);
}

1;
