
=head1 NAME

model3.pm

=head1 SYNOPSIS

Statistical library package to calculate the Loglikelihood ratio for 
trigrams. This package should be used with statistic.pl and rank.pl.

=head1 DESCRIPTION

The log-likelihood ratio measures the devitation between the observed data 
and what would be expected if <word1>, <word2>, and <word3> were  
independent. The higher the score, the less evidence there is in favor of  
concluding that the words are independent. 

=head1 AUTHORS

Ted Pedersen <tpederse@d.umn.edu>

Satanjeev Banerjee <banerjee@cs.cmu.edu>

Amruta Purandare <pura0010@d.umn.edu>

Bridget Thomson McInnes <bthomson@d.umn.edu>

=head1 BUGS

This measure currently only defined for trigram data stored in 3x3x3 
contingency table. 

=head1 SEE ALSO

Mailing List: http://groups.yahoo.com/ngram

=head1 COPYRIGHT

Copyright 2000-2004 by Ted Pedersen and Satanjeev Banerjee and Amruta 
Purandare

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

=cut

## this module written by Amruta Purandare, based on ll.pm by Satanjeev 
## Banerjee

package model3;

#  Make sure that model3model.pm is available in the PATH. First
#  we check in the directory you are running from, and then we
#  look through the system path. If the module is not found anywhere
#  then abort. 

my $module = "ll3model.pm"; my $modulename = "ll3model.pm";

if( !( -f $modulename ) ) {
    my $found = 0;
    #  Check each of the PATHS to see if the module is there
    foreach (split(/:/, $ENV{PATH})) {
 	$module = $_ . "/" . $modulename;
	if ( -f $module ) { $found = 1; last; }
    }
    # if still not found anywhere, quit!
    if ( ! $found ) { print "Could not find $modulename.\n"; exit; }
}

# Include the module into the current package.    

require $module;
require Exporter;

@ISA = qw ( Exporter );
@EXPORT = qw (initializeStatistic getStatisticName calculateStatistic errorCode errorString);

# function to set up various variables before the actual computation
# starts. also to check if we are being given bigrams, and if our
# frequency combinations are enough to do the computation
sub initializeStatistic
{    
    ll3model::initializeStatistic(@_);
}

# function to calculate the ll value!
sub calculateStatistic
{
    #  The parameters passed into calculateStatistic
    #  need to be passed into getObservedValues for it 
    #  to calculate the Observed values. If there is a 
    #  problem with the observed values, this function 
    #  will return, causing the measure to abort its 
    #  computations and return with a zero code. This is
    #  a required function to check the observed values.
    
    if( !( $check = ll3model::checkObservedValues(@_) ) ) {
	return(0);
    }

    #  Check the expected values. Also, the quotient of 
    #  the observed and expected value should not be less 
    #  than zero. If there is a problem with the expected 
    #  values, this function will return, causing the measure 
    #  to abort and return with a zero code.

     if( !( $check= ll3model::checkExpectedValues() ) ) {
	return(0);
    }

    #  Get the log likelihood values for each possible model
    #  ll1 - P(xy)P(z)
    #  ll2 - P(xz)P(y)
    #  ll3 - P(x)P(yz)
    #  ll4 - P(x)P(y)P(z)
    #  If there is a problem iwth the expected values, this 
    #  function will return, causing the measure to abort
    #  its computations and return with a zero code.

    if( !( ($ll1, $ll2, $ll3, $ll4) = ll3model::getLogLikelihoodValues() ) ) {
	return(0);
    }

    #  Return the loglikelihood score of the best fitting model
       
    $score = $ll1; my $model = 1;

    if($ll2 < $score) { $score = $ll2; $model = 2; }
    if($ll3 < $score) { $score = $ll3; $model = 3; }
    if($ll4 < $score) { $score = $ll4; $model = 4; }
        
    return $score;
    
}


# function to return the error code of the last operation and reset
# error code. useful if the error can be recovered from!
sub errorCode 
{ 
    return ll3model::errorCode();
}

# function to return the error message of the last operation and reset
# the message string. useful if error can be recovered from!
sub errorString
{
    return ll3model::errorString();
}

# function to return the name of this statistic
sub getStatisticName
{
    return "Loglikelihood3Modeling";
}

1;

